'From Squeak3.6alpha of ''17 March 2003'' [latest update: #5184] on 23 May 2003 at 11:38:08 am'!
"Change Set:		Closures-ajh
Date:			14 May 2004
Author:			Anthony Hannan <ajh18@cornell.edu>

This adds BlockClosures to the image.  Use the closure compiler in Compiler-ajh to create them.
"!

Object variableSubclass: #BlockClosure
	instanceVariableNames: 'method '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Kernel-Methods'!

!BlockClosure commentStamp: '<historical>' prior: 0!
A BlockClosure is a block of Smalltalk code (enclosed within []) that may be executed later by sending #valueWithArguments: (or one of its variants) to it.  A block can take arguments by specifying the names of the arguments in the beginning of the block, as in "[:arg1 :arg2 | ...]", and can have its own local temps, as in "[:arg1 |  | temp1 temp2 | ...]".  The block may reference variables outside its scope directly by name.  It also may return from its home context by using ^, otherwise, the value of the last statement is returned to the sender of valueWithArguments:.

Structure:

 method		CompiledMethod2
			Contains the block's code.  It has its own method separate from its home method.

 <indexable fields>  Objects
			Contains values of variables referenced outside its scope.  If a variable can change while being shared then it is wrapped in a SharedTempHolder, and the home context and each block references the holder and make changes to its contents.  If the block remotely returns (^) then its home context is also held in these fields.  If the block references self or one of its instance variables then self is also held in these fields.  See VarUse in the Compiler for how shared temps are determined.

See comment in class method #aboutClosures for more.
!

Inspector subclass: #ClosureInspector
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Tools-Debugger'!
Object subclass: #MethodClosure
	instanceVariableNames: 'receiver method '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Kernel-Contexts'!

!MethodClosure commentStamp: 'ajh 2/8/2003 23:14' prior: 0!
A MethodClosure is a method and receiver ready to be executed.  It is analogous to a MessageSend except the method is directly supplied instead of its selector.  This is useful for executing methods without installing them in the method dictionary (do-its).  It is called a closure because, like a BlockClosure, a MethodClosure holds the values (the receiver) it will need to execute.!

Object subclass: #PseudoPoolVariable
	instanceVariableNames: 'name getterBlock setterBlock '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Kernel-Contexts'!

!PseudoPoolVariable commentStamp: '<historical>' prior: 0!
The values of pool and global variables (traditionally Associations) are fetched by sending #poolValue and set by sending #setInPool: which send #poolValue:.  These sends are automatically added in by the Compiler (see PoolVarNode {code generation}).  So any object can act like a pool variable.  This class allows getter and setter blocks for poolValue and poolValue:.!

Object subclass: #SharedTempHolder
	instanceVariableNames: 'value '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Kernel-Contexts'!

!SharedTempHolder commentStamp: '<historical>' prior: 0!
Holds a changing temp that is shared among BlockClosures/homeContext.  The home method and block method(s) know to access the var indirectly through this holder.  See Compiler and VarUse for more on when these are used.

Structure:
 value		Object	- holds my value that may change!


!ProtoObject methodsFor: 'testing' stamp: 'ajh 12/4/2001 12:44'!
ifNotNil: ifNotNilBlock
	"Evaluate the block, unless I'm == nil (q.v.)"

	^ ifNotNilBlock valueWithPossibleArgs: {self}! !

!ProtoObject methodsFor: 'compiler prims' stamp: 'ajh 3/18/2003 23:21'!
getInstVar: index
	"The compiler generates this message when assigning to instance variables of objects that have been captured by a block closure.  Do not override or change this unless you know what your doing"

	<primitive: 73>
	^ self basicAt: index - self class instSize! !

!ProtoObject methodsFor: 'compiler prims' stamp: 'ajh 2/27/2003 13:15'!
remoteReturnTo: contextTag
	"Generated from compiler to return to the home of a block.  self is the value return and contextTag is the tag of the home context.  Do not override or change this unless you know what you are doing."

	contextTag return: self! !

!ProtoObject methodsFor: 'compiler prims' stamp: 'ajh 1/18/2003 13:17'!
setInPool: poolVariable
	"The compiler generates this message to place values in global variables.  Do not override or change this unless you know what your doing"

	poolVariable poolValue: self.
	^ self! !

!ProtoObject methodsFor: 'compiler prims' stamp: 'ajh 3/18/2003 23:23'!
setInstVar: index put: value 

	<primitive: 74>
	^ self basicAt: index - self class instSize put: value! !

!ProtoObject methodsFor: 'compiler prims' stamp: 'ajh 3/18/2003 23:22'!
storeIn: object instVar: index
	"The compiler generates this message when assigning to instance variables of objects that have been captured by a block closure.  Do not override or change this unless you know what your doing"

	object setInstVar: index put: self! !

!ProtoObject methodsFor: 'compiler prims' stamp: 'ajh 1/15/2003 21:23'!
storeInTempHolder: sharedTempHolder
	"The compiler generates this message to place temp values in holders so they can be shared amongst closures.  Do not override or change this unless you know what your doing"

	sharedTempHolder tempValue: self! !

!ProtoObject methodsFor: 'compiler prims' stamp: 'ajh 1/15/2003 21:23'!
wrapInTempHolder
	"The compiler generates this message to place temp values in holders so they can be shared amongst closures.  Do not override or change this unless you know what your doing"

	| holder |
	holder _ SharedTempHolder new.
	holder value: self.
	^ holder! !


!Object methodsFor: 'binding' stamp: 'ajh 2/6/2003 13:24'!
hasLiteralSuchThat: testBlock
	"This is the end of the imbedded structure path so return false."

	^ false! !

!Object methodsFor: 'binding' stamp: 'ajh 1/31/2003 12:04'!
hasLiteralThorough: literal
	"Answer true if literal is identical to any literal in this array, even if imbedded in further structures.  This is the end of the imbedded structure path so return false."

	^ false! !


!Array methodsFor: 'private' stamp: 'ajh 2/6/2003 13:22'!
hasLiteralSuchThat: testBlock

	| lit |
	1 to: self size do: [:index |
		(testBlock value: (lit _ self at: index)) ifTrue: [^ true].
		(lit hasLiteralSuchThat: testBlock) ifTrue: [^ true]].
	^ false! !

!Array methodsFor: 'private' stamp: 'ajh 1/31/2003 12:02'!
hasLiteralThorough: literal
	"Answer true if literal is identical to any literal in this array, even if imbedded in further array structures or closure methods"

	| lit |
	1 to: self size do: [:index |
		(lit _ self at: index) == literal ifTrue: [^ true].
		(lit hasLiteralThorough: literal) ifTrue: [^ true]].
	^ false! !


!Behavior methodsFor: 'accessing' stamp: 'ajh 3/13/2003 21:13'!
decompilerClass
	"Answer a decompiler class appropriate for compiled methods of this class."

	^ Preferences compileBlocksAsClosures
		ifTrue: [Decompiler2]
		ifFalse: [Decompiler]! !

!Behavior methodsFor: 'testing method dictionary' stamp: 'ajh 2/3/2003 17:40'!
thoroughWhichSelectorsReferTo: literal special: specialFlag byte: specialByte
	"Answer a set of selectors whose methods access the argument as a 
	literal. Dives into the compact literal notation, making it slow but 
	thorough "

	| who |
	who _ Set new.
	self selectorsAndMethodsDo:
		[:sel :method |
		((method hasLiteralThorough: literal) or: [specialFlag and: [method scanFor: specialByte]])
			ifTrue:
				[((literal isVariableBinding) not
					or: [method sendsToSuper not
					or: [method literals last ~= literal]])
						ifTrue: [who add: sel]]].
	^ who! !

!Behavior methodsFor: 'testing method dictionary' stamp: 'ajh 2/3/2003 17:36'!
whichSelectorsReferTo: literal special: specialFlag byte: specialByte
	"Answer a set of selectors whose methods access the argument as a literal."

	| who |
	who _ Set new.
	self selectorsAndMethodsDo: 
		[:sel :method |
		((method hasLiteralThorough: literal) or: [specialFlag and: [method scanFor: specialByte]])
			ifTrue:
				[((literal isVariableBinding) not
					or: [method sendsToSuper not
					or: [method literals last ~= literal]])
						ifTrue: [who add: sel]]].
	^ who! !


!BlockClosure methodsFor: 'initializing' stamp: 'ajh 5/28/2001 18:37'!
method: compiledMethod
	"compiledMethod will be the code I execute when I'm evaluated"

	method _ compiledMethod! !

!BlockClosure methodsFor: 'private' stamp: 'ajh 1/16/2002 18:32'!
copyForSaving
	"obsolete"! !

!BlockClosure methodsFor: 'private' stamp: 'ajh 1/16/2002 18:32'!
fixTemps
	"obsolete"! !

!BlockClosure methodsFor: 'private' stamp: 'ajh 1/31/2003 12:53'!
reentrant! !

!BlockClosure methodsFor: 'private' stamp: 'ajh 7/15/2001 16:13'!
valueError

	self error: 'Incompatible number of args'! !

!BlockClosure methodsFor: 'private' stamp: 'ajh 7/26/2002 11:47'!
valueUnpreemptively
	"Evaluate the receiver (block), without the possibility of preemption by higher priority processes. Use this facility VERY sparingly!!"
	"Think about using Block>>valueUninterruptably first, and think about using Semaphore>>critical: before that, and think about redesigning your application even before that!! 
	After you've done all that thinking, go right ahead and use it..."

	| activeProcess oldPriority result |
	activeProcess _ Processor activeProcess.
	oldPriority _ activeProcess priority.
	activeProcess priority: Processor highestPriority.
	result _ self ensure: [activeProcess priority: oldPriority].
	"Yield after restoring priority to give the preempted processes a chance to run"
	Processor yield.
	^result! !

!BlockClosure methodsFor: 'private' stamp: 'ajh 8/9/2002 13:11'!
veryDeepInner: deepCopier
	"Do not copy my method (which can be shared because CompiledMethod2 are basically treated as immutables) or my home context (MethodContexts are treated as immutables too)"

	super veryDeepInner: deepCopier.
	method _ method.
! !

!BlockClosure methodsFor: 'evaluating' stamp: 'ajh 1/13/2002 13:04'!
ifError: errorHandlerBlock
	"Evaluate the block represented by the receiver, and normally return it's value.  If an error occurs, the errorHandlerBlock is evaluated, and it's value is instead returned.  The errorHandlerBlock must accept zero, one, or two parameters (the error message and the receiver)."
	"Examples:
		[1 whatsUpDoc] ifError: [:err :rcvr | 'huh?'].
		[1 / 0] ifError: [:err :rcvr |
			'ZeroDivide' = err
				ifTrue: [Float infinity]
				ifFalse: [self error: err]]
"

	^ self on: Error do: [:ex |
		errorHandlerBlock valueWithPossibleArgs: {ex description. ex receiver}]! !

!BlockClosure methodsFor: 'evaluating'!
pvtEXECUTE
	^ t1! !

!BlockClosure methodsFor: 'evaluating' stamp: 'ajh 7/15/2001 15:57'!
timeToRun
	"Answer the number of milliseconds taken to execute this block."

	^ Time millisecondsToRun: self
! !

!BlockClosure methodsFor: 'evaluating' stamp: 'ajh 1/28/2003 12:22'!
value
	"Primitive. Evaluate the block represented by the receiver. Fail if the 
	block expects any arguments or if the block is already being executed. 
	Optional. No Lookup. See Object documentation whatIsAPrimitive."

	<primitive: 186>
	^ self valueWithArguments: #()! !

!BlockClosure methodsFor: 'evaluating' stamp: 'ajh 1/28/2003 12:22'!
value: arg 
	"Primitive. Evaluate the block represented by the receiver. Fail if the 
	block expects other than one argument or if the block is already being 
	executed. Optional. No Lookup. See Object documentation 
	whatIsAPrimitive."

	<primitive: 186>
	^ self valueWithArguments: (Array with: arg)! !

!BlockClosure methodsFor: 'evaluating' stamp: 'ajh 1/28/2003 12:26'!
value: arg1 value: arg2 
	"Primitive. Evaluate the block represented by the receiver. Fail if the 
	block expects other than two arguments or if the block is already being 
	executed. Optional. See Object documentation whatIsAPrimitive."

	<primitive: 186>
	^ self valueWithArguments: (Array with: arg1 with: arg2)! !

!BlockClosure methodsFor: 'evaluating' stamp: 'ajh 1/28/2003 12:26'!
value: arg1 value: arg2 value: arg3 
	"Primitive. Evaluate the block represented by the receiver. Fail if the 
	block expects other than three arguments or if the block is already being 
	executed. Optional. See Object documentation whatIsAPrimitive."

	<primitive: 186>
	^ self valueWithArguments: 
		(Array
			with: arg1
			with: arg2
			with: arg3)! !

!BlockClosure methodsFor: 'evaluating' stamp: 'ajh 1/28/2003 12:26'!
value: arg1 value: arg2 value: arg3 value: arg4 
	"Primitive. Evaluate the block represented by the receiver. Fail if the 
	block expects other than three arguments or if the block is already being 
	executed. Optional. See Object documentation whatIsAPrimitive."

	<primitive: 186>
	^ self valueWithArguments: 
		(Array
			with: arg1
			with: arg2
			with: arg3
			with: arg4)! !

!BlockClosure methodsFor: 'evaluating' stamp: 'ajh 1/28/2003 12:36'!
valueWithArguments: anArray 
	"Primitive. Evaluate the block represented by the receiver. The argument 
	is an Array whose elements are the arguments for the block. Fail if the 
	length of the Array is not the same as the the number of arguments that 
	the block was expecting. Fail if the block is already being executed. 
	Essential. See Object documentation whatIsAPrimitive."

	| dict index |
	<primitive: 187>
	#pvtEXECUTE flushCache.
	dict _ self class methodDict.
	index _ dict findElementOrNil: #pvtEXECUTE.
	dict methodArray at: index put: method.
	^ self perform: #pvtEXECUTE withArguments: anArray

	"Do not remove #pvtEXECUTE method just replace it and flush method cache.  Also avoid using an ensure block since that takes time also.  The primitive, if available, will execute the method directly without flushing the cache and adding to the method dictionary"
! !

!BlockClosure methodsFor: 'evaluating' stamp: 'ajh 1/28/2003 14:44'!
valueWithPossibleArgs: anArray 

	| n |
	(n _ self numArgs) = 0 ifTrue: [^ self value].
	n = anArray size ifTrue: [^ self valueWithArguments: anArray].
	^ self valueWithArguments: (n > anArray size
		ifTrue: [anArray, (Array new: n - anArray size)]
		ifFalse: [anArray copyFrom: 1 to: n])! !

!BlockClosure methodsFor: 'accessing' stamp: 'ajh 2/6/2003 13:24'!
hasLiteralSuchThat: testBlock

	(testBlock value: method) ifTrue: [^ true].
	^ method hasLiteralSuchThat: testBlock! !

!BlockClosure methodsFor: 'accessing' stamp: 'ajh 1/31/2003 16:59'!
hasLiteralThorough: literal
	"Answer true if literal is identical to any literal imbedded in my method"

	method == literal ifTrue: [^ true].
	^ method hasLiteralThorough: literal! !

!BlockClosure methodsFor: 'accessing' stamp: 'ajh 1/21/2003 13:16'!
isBlock

	^ true! !

!BlockClosure methodsFor: 'accessing' stamp: 'ajh 3/24/2003 01:41'!
longPrintOn: stream

	stream nextPutAll: self method blockNode decompileString.
	stream cr.
	self size = 0 ifTrue: [^ self].
	stream cr; nextPut: $".
	stream nextPutAll: self closureVarsAndValues.
	stream skip: -1; nextPut: $"; cr.
! !

!BlockClosure methodsFor: 'accessing' stamp: 'ajh 5/21/2001 14:01'!
method

	^ method! !

!BlockClosure methodsFor: 'accessing' stamp: 'ajh 5/28/2001 14:37'!
numArgs

	^ method numArgs! !

!BlockClosure methodsFor: 'accessing' stamp: 'ajh 9/10/2002 16:53'!
printOn: aStream 

	super printOn: aStream.
	aStream space; nextPutAll: self identityHashPrintString! !

!BlockClosure methodsFor: 'accessing' stamp: 'ajh 8/16/2002 15:54'!
vars

	| vars |
	vars _ Array new: self size.
	1 to: self size do: [:i | vars at: i put: (self at: i)].
	^ vars! !

!BlockClosure methodsFor: 'accessing' stamp: 'ajh 8/16/2002 15:50'!
varsDo: block

	1 to: self size do: [:i | block value: (self at: i)]! !

!BlockClosure methodsFor: 'controlling' stamp: 'ajh 7/15/2001 16:03'!
repeat
	"Evaluate the receiver repeatedly, ending only if the block explicitly returns."

	[self value. true] whileTrue! !

!BlockClosure methodsFor: 'controlling' stamp: 'ajh 7/15/2001 16:03'!
repeatWithGCIf: testBlock
	| ans |
	"run the receiver, and if testBlock returns true, garbage collect and run the receiver again"
	ans _ self value.
	(testBlock value: ans) ifTrue: [ Smalltalk garbageCollect. ans _ self value ].
	^ans! !

!BlockClosure methodsFor: 'controlling' stamp: 'ajh 7/15/2001 16:03'!
whileFalse
	"Ordinarily compiled in-line, and therefore not overridable.
	This is in case the message is sent to other than a literal block.
	Evaluate the receiver, as long as its value is false."
 
	^ [self value] whileFalse: []! !

!BlockClosure methodsFor: 'controlling' stamp: 'ajh 7/15/2001 16:03'!
whileFalse: aBlock 
	"Ordinarily compiled in-line, and therefore not overridable.
	This is in case the message is sent to other than a literal block.
	Evaluate the argument, aBlock, as long as the value of the receiver is false."

	^ [self value] whileFalse: [aBlock value]! !

!BlockClosure methodsFor: 'controlling' stamp: 'ajh 7/15/2001 16:03'!
whileTrue
	"Ordinarily compiled in-line, and therefore not overridable.
	This is in case the message is sent to other than a literal block.
	Evaluate the receiver, as long as its value is true."
 
	^ [self value] whileTrue: []! !

!BlockClosure methodsFor: 'controlling' stamp: 'ajh 7/15/2001 16:03'!
whileTrue: aBlock 
	"Ordinarily compiled in-line, and therefore not overridable.
	This is in case the message is sent to other than a literal block.
	Evaluate the argument, aBlock, as long as the value of the receiver is true."

	^ [self value] whileTrue: [aBlock value]! !

!BlockClosure methodsFor: 'exceptions' stamp: 'ajh 7/15/2001 16:14'!
assert
	self assert: self! !

!BlockClosure methodsFor: 'exceptions' stamp: 'ajh 1/21/2003 17:47'!
ensure: aBlock
	"Evaluate a termination block after evaluating the receiver, regardless of whether the receiver's evaluation completes."

	| returnValue |
	<primitive: 198>
	returnValue _ self value.
	"aBlock wasn't nil when execution of this method began; it is nil'd out by the unwind machinery, and that's how we know it's already been evaluated ... otherwise, obviously, it needs to be evaluated"
	aBlock == nil ifFalse: [aBlock value].
	^ returnValue! !

!BlockClosure methodsFor: 'exceptions' stamp: 'ajh 1/21/2003 17:50'!
ifCurtailed: aBlock
	"Evaluate the receiver with an abnormal termination action."

	<primitive: 198>
	^ self value! !

!BlockClosure methodsFor: 'exceptions' stamp: 'ajh 2/1/2003 00:30'!
on: exception do: handlerAction
	"Evaluate the receiver in the scope of an exception handler."

	| handlerActive |
	<primitive: 199>  "just a marker, fail and execute the following"
	handlerActive _ true.
	^ self value! !

!BlockClosure methodsFor: 'exceptions' stamp: 'ajh 1/31/2003 20:41'!
onDNU: selector do: handleBlock
	"Catch MessageNotUnderstood exceptions but only those of the given selector (DNU stands for doesNotUnderstand:)"

	^ self on: MessageNotUnderstood do: [:exception |
		exception message selector = selector
			ifTrue: [handleBlock valueWithPossibleArgs: {exception}]
			ifFalse: [exception pass]
	  ]! !

!BlockClosure methodsFor: 'exceptions' stamp: 'ajh 7/26/2002 11:49'!
valueUninterruptably
	"Prevent remote returns from escaping the sender.  Even attempts to terminate (unwind) this process will be halted and the process will resume here.  A terminate message is needed for every one of these in the sender chain to get the entire process unwound."

	^ self ifCurtailed: [^ self]! !

!BlockClosure methodsFor: 'scheduling' stamp: 'ajh 2/10/2003 14:24'!
asContext
	"Create a MethodContext that is ready to execute self.  Assumes self takes no args (if it does the args will be nil)"

	^ MethodContext 
		sender: nil
		receiver: self
		method: method
		arguments: #()! !

!BlockClosure methodsFor: 'scheduling' stamp: 'ajh 1/27/2003 18:51'!
callCC
	"Call with current continuation, ala Scheme.
	Evaluate self against a copy of the sender's call stack, which can be resumed later"

	^ self value: thisContext sender asContinuation! !

!BlockClosure methodsFor: 'scheduling' stamp: 'ajh 7/15/2001 16:03'!
fork
	"Create and schedule a Process running the code in the receiver."

	^ self newProcess resume! !

!BlockClosure methodsFor: 'scheduling' stamp: 'ajh 10/16/2002 11:14'!
forkAndWait
	"Suspend current process while self runs"

	| semaphore |
	semaphore _ Semaphore new.
	[self ensure: [semaphore signal]] fork.
	semaphore wait.
! !

!BlockClosure methodsFor: 'scheduling' stamp: 'ajh 9/29/2001 21:00'!
forkAt: priority 
	"Create and schedule a Process running the code in the receiver at the given priority. Answer the newly created process."

	^ self newProcess
		priority: priority;
		resume! !

!BlockClosure methodsFor: 'scheduling' stamp: 'ajh 2/10/2003 14:23'!
newProcess
	"Answer a Process running the code in the receiver. The process is not 
	scheduled."

	<primitive: 19> "Simulation guard"
	^ Process
		forContext: 
			[self value.
			Processor terminateActive] asContext
		priority: Processor activePriority! !

!BlockClosure methodsFor: 'scheduling' stamp: 'ajh 7/27/2002 12:26'!
simulate
	"Like run except interpret self using Smalltalk instead of VM.  It is much slower."

	^ self newProcess simulate! !

!BlockClosure methodsFor: 'inspecting' stamp: 'ajh 8/16/2002 15:51'!
closureAt: index
	"Answer the value of the closure variable at index"

	| v |
	v _ self at: index. 
	^ (v isMemberOf: SharedTempHolder)
		ifTrue: [v value]
		ifFalse: [v]! !

!BlockClosure methodsFor: 'inspecting' stamp: 'ajh 2/6/2002 14:36'!
closureAt: index put: value
	"Set the value of the closure variable at index"

	| v |
	v _ self at: index. 
	(v isKindOf: SharedTempHolder) ifTrue: [^ v value: value].
	^ self at: index put: value! !

!BlockClosure methodsFor: 'inspecting' stamp: 'ajh 3/16/2003 20:41'!
closureVarNames

	^ method blockNode freeNames! !

!BlockClosure methodsFor: 'inspecting' stamp: 'ajh 1/18/2002 11:24'!
closureVarsAndValues
	"Return a string of my vars with their current values"

	| aStream |
	aStream _ WriteStream on: (String new: 100).
	self closureVarNames doWithIndex: [:title :index |
		aStream nextPutAll: title; nextPut: $:; space; tab.
		(self closureAt: index) printOn: aStream.
		aStream cr].
	^ aStream contents! !

!BlockClosure methodsFor: 'inspecting' stamp: 'ajh 2/6/2003 13:27'!
hasMethodReturn
	"Answer whether the receiver has a return ('^') in its code."

	^ self method remoteReturns! !

!BlockClosure methodsFor: 'inspecting' stamp: 'ajh 2/3/2003 19:16'!
inspectorClass

	^ ClosureInspector! !

!BlockClosure methodsFor: 'inspecting' stamp: 'ajh 2/6/2003 12:48'!
searchImageForHomeMethod

	Smalltalk allObjectsDo: [:obj |
		obj class == CompiledMethod ifTrue: [
			(obj pointsTo: self) ifTrue: [^ obj searchImageForHomeMethod]
		] ifFalse: [obj class == BlockClosure ifTrue: [
			(obj method == self and: [obj size = 0])
				ifTrue: [^ obj searchImageForHomeMethod]
		]]
	].
	^ method  "must be a loner block method"! !

!BlockClosure methodsFor: 'Camp Smalltalk' stamp: 'rw 1/23/2002 00:27'!
sunitEnsure: aBlock

        ^self ensure: aBlock! !

!BlockClosure methodsFor: 'Camp Smalltalk' stamp: 'rw 1/23/2002 00:28'!
sunitOn: anException do: aHandlerBlock

        ^self on: anException do: aHandlerBlock! !

!BlockClosure methodsFor: 'comparing' stamp: 'ajh 10/4/2002 17:09'!
= other

	self class == other class ifFalse: [^ false].
	self size = other size ifFalse: [^ false].
	1 to: self size do: [:i |
		(self at: i) = (other at: i) ifFalse: [^ false].
	].
	^ method = other method! !

!BlockClosure methodsFor: 'comparing' stamp: 'ajh 10/4/2002 17:12'!
hash

	^ method hash! !


!Class methodsFor: 'compiling' stamp: 'ajh 6/18/2002 14:48'!
holder

	^ Smalltalk associationAt: name ifAbsent: [nil -> self]! !


!BlockClosure class methodsFor: 'documentation' stamp: 'ajh 3/13/2003 15:08'!
aboutClosures"

This package adds a closure compiler to the standard Squeak image (without replacing the existing compiler). A new 'general' preference called #compileBlocksAsClosures determines which compiler to use for new methods. When true the closure compiler is used, when false the existing compiler is used (default).

Etoy scripts/tiles do not work yet with the closure compiler so be sure to use them with the above preference set to false.

The closure compiler compiles non-inlined blocks as separate CompiledMethods. These block methods are held in the literals of the home method and sent the #createBlock... message at runtime to create BlockClosures. Outer temps referenced by the block are given as args to the #createBlock... message which adds them to the closure it creates, which consists of the method in its sole named field and the captured temps in its indexable fields. When #value... is sent to a BlockClosure its method is executed in a new MethodContext with the BlockClosure remaining as the receiver. The block method accesses captured temps in indexed fields as if they were named instance variables.

Captured temps that may change after being captured are wrapped in SharedTempHolders and all methods (home and block) access those values indirectly through the holders. Captured temps that may not change (most common) are placed directly in the closure, avoiding the creation and indirection of a holder.

BlockClosures that return to its home context hold the home context as one of its captured temps. Actually, to support safe continuations, it holds the tag of its home context. A ContextTag is a unique object for each context that stays the same for all copies of each context. When a BlockClosure attempts to return to its home context it searches the sender chain for the context that has the right ContextTag and returns to it. This works with copied contexts (continuations) as well as original contexts.

Thanks to captured temps and context tags, block closures are totally separate from their home context. They are reentrant and each activation has its own block-local temps.

BlockClosures are slower primarily because each activation creates a new context. BlockContexts are faster because they reuse the same context. An optimizing VM (Jitter, VI4) that stack allocates contexts should rectify this problem. For an example of its slowness (and general slowness of context activation) execute '[(1 to: 100000) do: [:i | i]] timeToRun' under each compiler (toggle the preference).

The above example will be really slow for the closure compiler until you generate interp.c and recompile your VM to include a couple of new primitives (186 and 187) added with the closure compiler. They implement BlockClosure>>value... directly. Without them the failure code has to add the closure method to the BlockClosure method dictionary then call #perform:.

The closure compiler is a total rewrite of the Smalltalk compiler using SmaCC parser generator and Refactory abstract systax tree.  For more details see the Compiler class comment.

Other General Enhancements

Context Control Enhancements - Methods like #return:, #restart, and #jump have been added to ContextPart for simpler control of the current thread.

Process Control Enhancements - Methods like #completeStep: have been added to Process for simpler control of suspended processes. 'completeStep: aContext' will resume the thread them jump back to the triggering thread when aContext get control back. This allows quick stepping while in the context of the suspended process (i.e. its exception handlers will be active instead of the triggering thread's).

Debugger Enhancements - The debugger uses the new process methods described above.

ContextTags - 'methodContext contextTag' will create a unique ContextTag object for that methodContext and hold it in its receiverMap instance variable which was previously unused. Copies of the methodContext will hold the same contextTag. Holding a contextTag instead of a context directly is safer in light of context copies (see continuations below). A context for a context tag is found in the current thread just be traversing the senders until one is found with that contextTag.

Exceptions Refactored - Exceptions now hold initialContext and handlerContext tags instead of the actual contexts so contexts can be copied safely (described above). Also, searching for handler contexts has been simplified utilizing Context control behavior described above.

Continuations - 'Continuation current' copies thisContext stack and wraps it in a Continuation object. Also, '[:cc | ...] callCC' copies the sender stack into a Continuation and returning it. Evaluating a continuation copies the stack again and resumes that copy. The current process stack is abandoned. When a BlockContext on the stack is copied it points to copied home context if it is also on the stack.

Miscellaneous - Changeset 1 contains general additions to basic objects, collections, and streams, that are used by the rest of changes described above.

Test Cases

Thanks to Boris Gaertner and Rob Withers for writing closure test cases (BlockClosuresTestCase). And thanks to Avi Bryant for writing continuation test cases (ContinuationTest and AmbTest).

"! !

!BlockClosure class methodsFor: 'compiling' stamp: 'ajh 8/9/2002 01:52'!
varOffset

	^ self instSize! !


!ClosureInspector methodsFor: 'accessing' stamp: 'ajh 1/18/2002 12:29'!
doItContext
	"Use my closure as the context to get closure variables"

	^ object! !

!ClosureInspector methodsFor: 'accessing' stamp: 'ajh 9/14/2002 09:51'!
doItReceiver
	"My doItContext will take precedence"

	^ nil! !

!ClosureInspector methodsFor: 'accessing' stamp: 'ajh 2/4/2003 13:21'!
fieldList 
	"Refer to the comment in Inspector|fieldList."

	object == nil ifTrue: [^ #('thisClosure')].
	^ #('thisClosure' 'block code' 'all free vars'), object closureVarNames! !

!ClosureInspector methodsFor: 'selecting' stamp: 'ajh 3/17/2003 12:29'!
contentsIsString
	"Hacked so contents empty when deselected and = long printString when item 2"

	^ (selectionIndex between: 2 and: 3) | (selectionIndex = 0)! !

!ClosureInspector methodsFor: 'selecting' stamp: 'ajh 2/9/2003 16:19'!
replaceSelectionValue: anObject 
	"Refer to the comment in Inspector|replaceSelectionValue:."

	selectionIndex = 1 ifTrue: [^ object].
	^ object closureAt: selectionIndex - 3 put: anObject! !

!ClosureInspector methodsFor: 'selecting' stamp: 'ajh 2/4/2003 13:20'!
selection 
	"Refer to the comment in Inspector|selection."

	selectionIndex = 0 ifTrue:[^ ''].
	selectionIndex = 1 ifTrue: [^ object].
	selectionIndex = 2 ifTrue: [^ object method blockNode decompileString].
	selectionIndex = 3 ifTrue: [^ object closureVarsAndValues].
	^ object closureAt: selectionIndex - 3! !


!CompiledMethod methodsFor: 'initialize-release' stamp: 'ajh 1/22/2003 16:26'!
isClosureCompiled: bool
	"Use the sign bit in the header to mark methods that have been compiled using the new closure compiler (Parser2)."

	bool
		ifTrue: [self objectAt: 1 put: (self header bitOr: 1 << 30) as31BitSmallInt]
		ifFalse: [self error: 'Don''t send if not closure compiled']! !

!CompiledMethod methodsFor: 'accessing' stamp: 'ajh 3/10/2003 14:02'!
method
	"polymorphic with closure"

	^ self! !

!CompiledMethod methodsFor: 'literals' stamp: 'ajh 2/6/2003 13:25'!
hasLiteralSuchThat: testBlock
	"Answer true if testBlock returns true for any literal in this method, even if imbedded."

	| lit |
	2 to: self numLiterals + 1 do: 
		[:index | 
		(testBlock value: (lit _ self objectAt: index)) ifTrue: [^ true].
		(lit hasLiteralSuchThat: testBlock) ifTrue: [^ true]].
	^ false! !

!CompiledMethod methodsFor: 'literals' stamp: 'ajh 1/31/2003 12:05'!
hasLiteralThorough: literal
	"Answer true if any literal in this method is literal,
	even if embedded in array structure."

	| lit |
	2 to: self numLiterals + 1 do: 
		[:index | 
		(lit _ self objectAt: index) == literal ifTrue: [^ true].
		(lit hasLiteralThorough: literal) ifTrue: [^ true]].
	^ false! !

!CompiledMethod methodsFor: 'scanning' stamp: 'ajh 2/6/2003 13:27'!
remoteReturns
	"For closure methods only"

	^ self messages includes: #remoteReturnTo:! !

!CompiledMethod methodsFor: 'decompiling' stamp: 'ajh 7/14/2001 12:34'!
primitiveNode

	| primNode n |
	primNode _ PrimitiveNode new num: (n _ self primitive).
	(n = 117 or: [n = 120]) ifTrue: [
		primNode spec: (self literalAt: 1)].
	^ primNode! !

!CompiledMethod methodsFor: 'decompiling' stamp: 'ajh 2/6/2003 12:48'!
searchImageForHomeMethod

	Smalltalk allObjectsDo: [:obj |
		obj class == CompiledMethod ifTrue: [
			(obj pointsTo: self) ifTrue: [^ obj searchImageForHomeMethod]
		] ifFalse: [obj class == BlockClosure ifTrue: [
			(obj method == self and: [obj size = 0])
				ifTrue: [^ obj searchImageForHomeMethod]
		]]
	].
	^ self  "must be a loner block method"! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:29'!
createBlock
	"Generate by the compiler to create a block closure with self as its method and no captured temps"

	^ BlockClosure new method: self! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 2/10/2003 15:02'!
createBlockWithVars: array
	"Generate by the compiler to create a block closure with self as its method and array elements as its captured temps"

	| block |
	block _ (BlockClosure new: array size) method: self.
	1 to: array size do: [:i | block at: i put: (array at: i)].
	^ block! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:23'!
createBlockt: t1
	"Generate by the compiler to create a block closure with self as its method, and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 1) method: self;
		at: 1 put: t1;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:22'!
createBlockt: t1 t: t2
	"Generate by the compiler to create a block closure with self as its method, and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 2) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:22'!
createBlockt: t1 t: t2 t: t3
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 3) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:23'!
createBlockt: t1 t: t2 t: t3 t: t4
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 4) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:23'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 5) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:24'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5 t: t6
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 6) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		at: 6 put: t6;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:24'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5 t: t6 t: t7
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 7) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		at: 6 put: t6;
		at: 7 put: t7;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:24'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5 t: t6 t: t7 t: t8
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 8) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		at: 6 put: t6;
		at: 7 put: t7;
		at: 8 put: t8;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:25'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5 t: t6 t: t7 t: t8 t: t9
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 9) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		at: 6 put: t6;
		at: 7 put: t7;
		at: 8 put: t8;
		at: 9 put: t9;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:25'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5 t: t6 t: t7 t: t8 t: t9 t: t10
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 10) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		at: 6 put: t6;
		at: 7 put: t7;
		at: 8 put: t8;
		at: 9 put: t9;
		at: 10 put: t10;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:25'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5 t: t6 t: t7 t: t8 t: t9 t: t10 t: t11
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 11) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		at: 6 put: t6;
		at: 7 put: t7;
		at: 8 put: t8;
		at: 9 put: t9;
		at: 10 put: t10;
		at: 11 put: t11;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:26'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5 t: t6 t: t7 t: t8 t: t9 t: t10 t: t11 t: t12
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 12) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		at: 6 put: t6;
		at: 7 put: t7;
		at: 8 put: t8;
		at: 9 put: t9;
		at: 10 put: t10;
		at: 11 put: t11;
		at: 12 put: t12;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:26'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5 t: t6 t: t7 t: t8 t: t9 t: t10 t: t11 t: t12 t: t13
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 13) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		at: 6 put: t6;
		at: 7 put: t7;
		at: 8 put: t8;
		at: 9 put: t9;
		at: 10 put: t10;
		at: 11 put: t11;
		at: 12 put: t12;
		at: 13 put: t13;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:26'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5 t: t6 t: t7 t: t8 t: t9 t: t10 t: t11 t: t12 t: t13 t: t14
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 14) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		at: 6 put: t6;
		at: 7 put: t7;
		at: 8 put: t8;
		at: 9 put: t9;
		at: 10 put: t10;
		at: 11 put: t11;
		at: 12 put: t12;
		at: 13 put: t13;
		at: 14 put: t14;
		yourself! !

!CompiledMethod methodsFor: 'closure creation' stamp: 'ajh 1/15/2003 22:27'!
createBlockt: t1 t: t2 t: t3 t: t4 t: t5 t: t6 t: t7 t: t8 t: t9 t: t10 t: t11 t: t12 t: t13 t: t14 t: t15
	"Generate by the compiler to create a block closure with self as its method and t1, t2, ... as its captured outer temps"

	^ (BlockClosure new: 15) method: self;
		at: 1 put: t1;
		at: 2 put: t2;
		at: 3 put: t3;
		at: 4 put: t4;
		at: 5 put: t5;
		at: 6 put: t6;
		at: 7 put: t7;
		at: 8 put: t8;
		at: 9 put: t9;
		at: 10 put: t10;
		at: 11 put: t11;
		at: 12 put: t12;
		at: 13 put: t13;
		at: 14 put: t14;
		at: 15 put: t15;
		yourself! !


!ContextPart methodsFor: 'accessing' stamp: 'ajh 2/3/2003 21:39'!
temp: index
	"Answer the value of the temporary variable whose index is the 
	argument, index.  The first argument is temp 1"

	^ self subclassResponsibility! !

!ContextPart methodsFor: 'accessing' stamp: 'ajh 2/3/2003 21:40'!
temp: index put: value 
	"Store the argument, value, as the temporary variable whose index is the 
	argument, index.  The first arg of self is temp 1"

	^ self subclassResponsibility! !

!ContextPart methodsFor: 'accessing' stamp: 'ajh 2/3/2003 21:45'!
tempAt: index
	"Get the temp value at index.  Make SharedTempHolders transparent."

	| var |
	var _ self temp: index.
	^ (var isMemberOf: SharedTempHolder)
		ifTrue: [var value]
		ifFalse: [var]! !

!ContextPart methodsFor: 'accessing' stamp: 'ajh 2/3/2003 21:45'!
tempAt: index put: val
	"Set the temp at index.  Make SharedTempHolders transparent."

	| var |
	var _ self temp: index.
	^ (var isMemberOf: SharedTempHolder)
		ifTrue: [var value: val]
		ifFalse: [self temp: index put: val]! !

!ContextPart methodsFor: 'system simulation' stamp: 'ajh 4/15/2003 10:01'!
runSimulated: aBlock contextAtEachStep: block2
	"Simulate the execution of the argument, aBlock, until it ends. aBlock 
	MUST NOT contain an '^'. Evaluate block2 with the current context 
	prior to each instruction executed. Answer the simulated value of aBlock."

	| current |
	(aBlock isBlock and: [aBlock hasMethodReturn])
		ifTrue: [self error: 'simulation of blocks with ^ can run loose'].
	current _ aBlock asContext.
	current privSender: self.
	[current == self]
		whileFalse:
			[block2 value: current.
			current _ current step].
	^ self pop! !

!ContextPart methodsFor: 'private' stamp: 'ajh 1/31/2003 15:07'!
doPrimitive: primitiveIndex method: meth receiver: receiver args: arguments 
	"Simulate a primitive method whose index is primitiveIndex.  The
	simulated receiver and arguments are given as arguments to this message."

	| value |
	<primitive: 19> "Simulation guard"
	"If successful, push result and return resuming context,
		else ^ PrimitiveFailToken"
	(primitiveIndex = 19) ifTrue:[
		Debugger 
			openContext: self
			label:'Code simulation error'
			contents: nil].

	(primitiveIndex = 80 and: [receiver isKindOf: ContextPart])
		ifTrue: [^self push: ((BlockContext newForMethod: receiver home method)
						home: receiver home
						startpc: pc + 2
						nargs: (arguments at: 1))].
	(primitiveIndex = 81 and: [receiver isMemberOf: BlockContext])
		ifTrue: [^receiver pushArgs: arguments from: self].
	primitiveIndex = 83 "afr 9/11/1998 19:50"
		ifTrue: [^ self send: arguments first to: receiver
					with: arguments allButFirst
					super: false].
	primitiveIndex = 84 "afr 9/11/1998 19:50"
		ifTrue: [^ self send: arguments first to: receiver
					with: (arguments at: 2)
					super: false].
	primitiveIndex = 186 ifTrue: [ "closure value"
		| m |
		m _ receiver method.
		arguments size = m numArgs ifFalse: [^ PrimitiveFailToken].
		^ self activateMethod: m
			withArgs: arguments
			receiver: receiver
			class: receiver class].
	primitiveIndex = 187 ifTrue: [ "closure valueWithArguments:"
		| m args |
		m _ receiver method.
		args _ arguments first.
		args size = m numArgs ifFalse: [^ PrimitiveFailToken].
		^ self activateMethod: m
			withArgs: args
			receiver: receiver
			class: receiver class].
	arguments size > 6 ifTrue: [^ PrimitiveFailToken].
	primitiveIndex = 117 
		ifTrue:[value _ self tryNamedPrimitiveIn: meth for: receiver withArgs: arguments]
		ifFalse:[value _ receiver tryPrimitive: primitiveIndex withArgs: arguments].
	value == PrimitiveFailToken
		ifTrue: [^ PrimitiveFailToken]
		ifFalse: [^ self push: value]! !

!ContextPart methodsFor: 'query' stamp: 'ajh 1/31/2003 21:53'!
isClosureContext

	^ false! !


!BlockContext methodsFor: 'accessing' stamp: 'ajh 2/3/2003 21:40'!
temp: index 
	"Refer to the comment in ContextPart|temp:."

	^ home at: index! !

!BlockContext methodsFor: 'accessing' stamp: 'ajh 2/3/2003 21:40'!
temp: index put: value 
	"Refer to the comment in ContextPart|temp:put:."

	^ home at: index put: value! !

!BlockContext methodsFor: 'scheduling' stamp: 'ajh 2/10/2003 14:23'!
asContext

	^ self! !

!BlockContext methodsFor: 'scheduling' stamp: 'ajh 2/10/2003 14:25'!
newProcess
	"Answer a Process running the code in the receiver. The process is not 
	scheduled."
	<primitive: 19> "Simulation guard"
	^Process
		forContext: 
			[self value.
			Processor terminateActive] asContext
		priority: Processor activePriority! !

!BlockContext methodsFor: 'scheduling' stamp: 'ajh 2/10/2003 14:25'!
newProcessWith: anArray 
	"Answer a Process running the code in the receiver. The receiver's block 
	arguments are bound to the contents of the argument, anArray. The 
	process is not scheduled."
	<primitive: 19> "Simulation guard"
	^Process
		forContext: 
			[self valueWithArguments: anArray.
			Processor terminateActive] asContext
		priority: Processor activePriority! !


!LookupKey methodsFor: 'accessing' stamp: 'ajh 9/12/2002 12:04'!
canAssign

	^ true! !

!LookupKey methodsFor: 'accessing' stamp: 'ajh 9/12/2002 02:27'!
poolName

	^ self key! !

!LookupKey methodsFor: 'accessing' stamp: 'ajh 7/21/2002 13:15'!
poolValue

	^ self value! !

!LookupKey methodsFor: 'accessing' stamp: 'ajh 7/21/2002 13:16'!
poolValue: obj

	self value: obj.
	^ obj! !


!Association methodsFor: 'accessing' stamp: 'ajh 9/19/2002 23:58'!
poolValue

	^ value! !


!MessageTally methodsFor: 'initialize-release' stamp: 'ajh 2/9/2003 20:21'!
spyEvery: millisecs on: aBlock 
	"Create a spy and spy on the given block at the specified rate."

	| myDelay value startTime time0 |
	(aBlock isMemberOf: BlockContext)
		ifTrue: [self class: aBlock receiver class method: aBlock method]
		ifFalse: [self class: aBlock class method: aBlock method].
	"set up the probe"
	ObservedProcess _ Processor activeProcess.
	myDelay := Delay forMilliseconds: millisecs.
	time0 := Time millisecondClockValue.
	gcStats _ Smalltalk getVMParameters.
	Timer :=
		[[true] whileTrue: 
			[startTime := Time millisecondClockValue.
			myDelay wait.
			self tally: ObservedProcess suspendedContext
				"tally can be > 1 if ran a long primitive"
				by: (Time millisecondClockValue - startTime) // millisecs].
		nil] newProcess.
	Timer priority: Processor userInterruptPriority.
		"activate the probe and evaluate the block"
	Timer resume.
	value := aBlock value.
	"Collect gc statistics"
	Smalltalk getVMParameters keysAndValuesDo:
		[:idx :gcVal| gcStats at: idx put: (gcVal - gcStats at: idx)].
		"cancel the probe and return the value"
	Timer terminate.
	time := Time millisecondClockValue - time0.
	^value! !


!MessageTally class methodsFor: 'spying' stamp: 'ajh 4/15/2003 09:55'!
tallySendsTo: receiver inBlock: aBlock showTree: treeOption
	"MessageTally tallySends: [3.14159 printString]"
	"This method uses the simulator to count the number of calls on each method
	invoked in evaluating aBlock. If receiver is not nil, then only sends
	to that receiver are tallied.
	Results are presented as leaves, sorted by frequency,
	preceded, optionally, by the whole tree."
	| prev tallies startTime totalTime |
	startTime _ Time millisecondClockValue.
	tallies _ MessageTally new class: aBlock class
							method: aBlock method.
	prev _ aBlock.
	thisContext sender
		runSimulated: aBlock
		contextAtEachStep:
			[:current |
			current == prev ifFalse: 
				["call or return"
				prev sender == nil ifFalse: 
					["call only"
					(receiver == nil or: [current receiver == receiver])
						ifTrue: [tallies tally: current by: 1]].
				prev _ current]].

	totalTime _ Time millisecondClockValue - startTime // 1000.0 roundTo: 0.01.
	(StringHolder new contents:
		(String streamContents:
			[:s |
			s nextPutAll: 'This simulation took ' , totalTime printString
							, ' seconds.'; cr.
			treeOption
				ifTrue: [tallies fullPrintOn: s tallyExact: true orThreshold: 0]
				ifFalse: [tallies leavesPrintOn: s tallyExact: true orThreshold: 0].
			tallies close]))
		openLabel: 'Spy Results'! !


!Metaclass methodsFor: 'compiling' stamp: 'ajh 6/18/2002 14:48'!
holder

	^ nil -> self! !


!MethodClosure methodsFor: 'as yet unclassified' stamp: 'ajh 2/8/2003 23:01'!
method: compiledMethod

	method _ compiledMethod! !

!MethodClosure methodsFor: 'as yet unclassified' stamp: 'ajh 2/8/2003 23:01'!
receiver: obj

	receiver _ obj! !

!MethodClosure methodsFor: 'as yet unclassified' stamp: 'ajh 2/8/2003 22:53'!
value

	^ receiver withArgs: #() executeMethod: method! !


!MethodContext methodsFor: 'accessing' stamp: 'ajh 1/31/2003 21:53'!
isClosureContext

	^ self isExecutingBlock! !

!MethodContext methodsFor: 'accessing' stamp: 'ajh 2/3/2003 21:39'!
temp: index 
	"Refer to the comment in ContextPart|temp:."

	^ self at: index! !

!MethodContext methodsFor: 'accessing' stamp: 'ajh 2/3/2003 21:40'!
temp: index put: value 
	"Refer to the comment in ContextPart|temp:put:."

	^ self at: index put: value! !

!MethodContext methodsFor: 'private' stamp: 'ajh 4/15/2003 10:03'!
asContext

	^ self! !


!PseudoPoolVariable methodsFor: 'as yet unclassified' stamp: 'ajh 9/12/2002 12:08'!
canAssign

	^ setterBlock notNil! !

!PseudoPoolVariable methodsFor: 'as yet unclassified' stamp: 'ajh 9/12/2002 03:01'!
getter: block

	getterBlock _ block! !

!PseudoPoolVariable methodsFor: 'as yet unclassified' stamp: 'ajh 9/12/2002 03:00'!
name: string

	name _ string! !

!PseudoPoolVariable methodsFor: 'as yet unclassified' stamp: 'ajh 9/12/2002 03:01'!
poolName

	^ name! !

!PseudoPoolVariable methodsFor: 'as yet unclassified' stamp: 'ajh 9/12/2002 03:01'!
poolValue

	^ getterBlock value! !

!PseudoPoolVariable methodsFor: 'as yet unclassified' stamp: 'ajh 9/12/2002 03:02'!
poolValue: obj

	setterBlock value: obj! !

!PseudoPoolVariable methodsFor: 'as yet unclassified' stamp: 'ajh 9/12/2002 03:01'!
setter: block

	setterBlock _ block! !


!ReadOnlyVariableBinding methodsFor: 'accessing' stamp: 'ajh 9/12/2002 12:06'!
canAssign

	^ false! !

!ReadOnlyVariableBinding methodsFor: 'accessing' stamp: 'ajh 9/19/2002 23:58'!
poolValue

	^ value! !


!SharedTempHolder methodsFor: 'as yet unclassified' stamp: 'ajh 9/17/2001 21:54'!
printOn: stream

	super printOn: stream.
	stream nextPut: $(.
	value printOn: stream.
	stream nextPut: $).
! !

!SharedTempHolder methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 20:25'!
tempHolderValue

	^ value! !

!SharedTempHolder methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 20:07'!
tempValue

	^ value! !

!SharedTempHolder methodsFor: 'as yet unclassified' stamp: 'ajh 1/15/2003 20:07'!
tempValue: obj

	value _ obj! !

!SharedTempHolder methodsFor: 'as yet unclassified' stamp: 'ajh 9/15/2001 19:50'!
value

	^ value! !

!SharedTempHolder methodsFor: 'as yet unclassified' stamp: 'ajh 9/15/2001 19:50'!
value: obj

	^ value _ obj! !


!SharedTempHolder class methodsFor: 'as yet unclassified' stamp: 'ajh 6/5/2002 22:55'!
valueSlotIndex
	"Return the slot index of 'value' in my instances"

	^ 1! !

!SharedTempHolder class methodsFor: 'as yet unclassified' stamp: 'ajh 9/26/2001 09:37'!
with: value

	^ self new
		value: value;
		yourself! !


!SystemDictionary methodsFor: 'housekeeping' stamp: 'ajh 4/15/2003 09:59'!
macroBenchmark3   "Smalltalk macroBenchmark3"
	| testBlock tallies prev receiver |
	"Runs the stepping simulator with the messageTally tree (like tallySends)."
	testBlock _
		['Running the context step simulator'
			displayProgressAt: Sensor cursorPoint
			from: 0 to: 200
			during:
				[:bar |
				1 to: 200 do:
				[:x | bar value: x.
				Float pi printString.
				15 factorial printString]]].
	tallies _ MessageTally new class: testBlock class
							method: testBlock method.
	receiver _ nil.
	prev _ testBlock asContext.
	thisContext sender
		runSimulated: prev
		contextAtEachStep:
			[:current |
			current == prev ifFalse: 
				["call or return"
				prev sender == nil ifFalse: 
					["call only"
					(receiver == nil or: [current receiver == receiver])
						ifTrue: [tallies tally: current by: 1]].
				prev _ current]].
! !

MethodContext removeSelector: #tempAt:!
MethodContext removeSelector: #tempAt:put:!
BlockContext removeSelector: #tempAt:!
BlockContext removeSelector: #tempAt:put:!
